from functools import partial
from typing import Callable

import nltk
from krovetzstemmer import Stemmer
from nltk.stem import (
    ARLSTem,
    ARLSTem2,
    Cistem,
    ISRIStemmer,
    LancasterStemmer,
    PorterStemmer,
    RSLPStemmer,
    SnowballStemmer,
)

nltk.download("rslp", quiet=True)

stemmers_dict = {
    "krovetz": partial(Stemmer()),
    "porter": partial(PorterStemmer().stem),
    "lancaster": partial(LancasterStemmer().stem),
    "arlstem": partial(ARLSTem().stem),  # Arabic
    "arlstem2": partial(ARLSTem2().stem),  # Arabic
    "cistem": partial(Cistem().stem),  # German
    "isri": partial(ISRIStemmer().stem),  # Arabic
    "rslp": partial(RSLPStemmer().stem),  # Portuguese
    "arabic": partial(SnowballStemmer("arabic").stem),
    "snowball-arabic": partial(SnowballStemmer("arabic").stem),
    "danish": partial(SnowballStemmer("danish").stem),
    "snowball-danish": partial(SnowballStemmer("danish").stem),
    "dutch": partial(SnowballStemmer("dutch").stem),
    "snowball-dutch": partial(SnowballStemmer("dutch").stem),
    "english": partial(SnowballStemmer("english").stem),
    "snowball-english": partial(SnowballStemmer("english").stem),
    "finnish": partial(SnowballStemmer("finnish").stem),
    "snowball-finnish": partial(SnowballStemmer("finnish").stem),
    "french": partial(SnowballStemmer("french").stem),
    "snowball-french": partial(SnowballStemmer("french").stem),
    "german": partial(SnowballStemmer("german").stem),
    "snowball-german": partial(SnowballStemmer("german").stem),
    "hungarian": partial(SnowballStemmer("hungarian").stem),
    "snowball-hungarian": partial(SnowballStemmer("hungarian").stem),
    "italian": partial(SnowballStemmer("italian").stem),
    "snowball-italian": partial(SnowballStemmer("italian").stem),
    "norwegian": partial(SnowballStemmer("norwegian").stem),
    "snowball-norwegian": partial(SnowballStemmer("norwegian").stem),
    "porter": partial(SnowballStemmer("porter").stem),
    "snowball-porter": partial(SnowballStemmer("porter").stem),
    "portuguese": partial(SnowballStemmer("portuguese").stem),
    "snowball-portuguese": partial(SnowballStemmer("portuguese").stem),
    "romanian": partial(SnowballStemmer("romanian").stem),
    "snowball-romanian": partial(SnowballStemmer("romanian").stem),
    "russian": partial(SnowballStemmer("russian").stem),
    "snowball-russian": partial(SnowballStemmer("russian").stem),
    "spanish": partial(SnowballStemmer("spanish").stem),
    "snowball-spanish": partial(SnowballStemmer("spanish").stem),
    "swedish": partial(SnowballStemmer("swedish").stem),
    "snowball-swedish": partial(SnowballStemmer("swedish").stem),
}


def krovetz_f(x: str) -> str:
    return stemmers_dict["krovetz"](x)


def porter_f(x: str) -> str:
    return stemmers_dict["porter"](x)


def lancaster_f(x: str) -> str:
    return stemmers_dict["lancaster"](x)


def arlstem_f(x: str) -> str:
    return stemmers_dict["arlstem"](x)


def arlstem2_f(x: str) -> str:
    return stemmers_dict["arlstem2"](x)


def cistem_f(x: str) -> str:
    return stemmers_dict["cistem"](x)


def isri_f(x: str) -> str:
    return stemmers_dict["isri"](x)


def rslp_f(x: str) -> str:
    return stemmers_dict["rslp"](x)


def arabic_f(x: str) -> str:
    return stemmers_dict["arabic"](x)


def danish_f(x: str) -> str:
    return stemmers_dict["danish"](x)


def dutch_f(x: str) -> str:
    return stemmers_dict["dutch"](x)


def english_f(x: str) -> str:
    return stemmers_dict["english"](x)


def finnish_f(x: str) -> str:
    return stemmers_dict["finnish"](x)


def french_f(x: str) -> str:
    return stemmers_dict["french"](x)


def german_f(x: str) -> str:
    return stemmers_dict["german"](x)


def hungarian_f(x: str) -> str:
    return stemmers_dict["hungarian"](x)


def italian_f(x: str) -> str:
    return stemmers_dict["italian"](x)


def norwegian_f(x: str) -> str:
    return stemmers_dict["norwegian"](x)


def porter_f(x: str) -> str:
    return stemmers_dict["porter"](x)


def portuguese_f(x: str) -> str:
    return stemmers_dict["portuguese"](x)


def romanian_f(x: str) -> str:
    return stemmers_dict["romanian"](x)


def russian_f(x: str) -> str:
    return stemmers_dict["russian"](x)


def spanish_f(x: str) -> str:
    return stemmers_dict["spanish"](x)


def swedish_f(x: str) -> str:
    return stemmers_dict["swedish"](x)


def wordnet_f(x: str) -> str:
    return stemmers_dict["wordnet"](x)


stemmers_f_dict = {
    "krovetz": krovetz_f,
    "porter": porter_f,
    "lancaster": lancaster_f,
    "arlstem": arlstem_f,
    "arlstem2": arlstem2_f,
    "cistem": cistem_f,
    "isri": isri_f,
    "rslp": rslp_f,
    "arabic": arabic_f,
    "danish": danish_f,
    "dutch": dutch_f,
    "english": english_f,
    "finnish": finnish_f,
    "french": french_f,
    "german": german_f,
    "hungarian": hungarian_f,
    "italian": italian_f,
    "norwegian": norwegian_f,
    "porter": porter_f,
    "portuguese": portuguese_f,
    "romanian": romanian_f,
    "russian": russian_f,
    "spanish": spanish_f,
    "swedish": swedish_f,
    "wordnet": wordnet_f,
}


def get_stemmer(stemmer: str) -> Callable:
    assert (
        stemmer.lower() in stemmers_f_dict
    ), f"Stemmer {stemmer} not supported."
    return stemmers_f_dict[stemmer.lower()]



import os
import shutil
import string
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path
from typing import Callable, Dict, List, Union

import numba as nb
import numpy as np
import orjson
from indxr import Indxr
from numba.typed import List as TypedList
from oneliner_utils import read_csv, read_jsonl
from rich.progress import track

from .autotune import tune_bm25
from .bm25 import bm25, bm25_multi
from .stemmers import get_stemmer
from .stopwords import get_stopwords
from .tokenizers import get_tokenizer


def home_path():
    p = Path(Path.home() / ".collections")
    p.mkdir(parents=True, exist_ok=True)
    return p


def prepro(x: str, tokenizer: Callable, sw_list: list, stemmer: Callable):
    x = x.lower()
    x = x.translate(str.maketrans("", "", string.punctuation))  # Remove punct
    x = tokenizer(x)
    x = [t for t in x if t not in sw_list]

    return [stemmer(t) for t in x]


def prepro_caller(input):
    return prepro(*input)


class SearchEngine:
    def __init__(
        self,
        index_name: str = "new-index",
        min_term_freq: int = 1,
        tokenizer: Union[str, Callable] = "whitespace",
        stemmer: Union[str, Callable] = "english",
        sw_list: Union[str, List] = "english",
    ):
        assert min_term_freq > 0, "`min_term_freq` must be greater than zero."
        self.init_args = {
            "index_name": index_name,
            "tokenizer": tokenizer,
            "stemmer": stemmer,
            "sw_list": sw_list,
        }
        self.min_term_freq = min_term_freq

        self.index_name = index_name
        self.index_path = home_path() / index_name
        self.index_path.mkdir(parents=True, exist_ok=True)

        if type(tokenizer) is str:
            self.tokenizer = get_tokenizer(tokenizer)
        elif callable(tokenizer):
            self.tokenizer = tokenizer
        elif tokenizer is False:
            self.tokenizer = lambda x: x
        else:
            raise (NotImplementedError)

        if type(stemmer) is str:
            self.stemmer = get_stemmer(stemmer)
        elif callable(stemmer):
            self.stemmer = stemmer
        elif stemmer is False:
            self.stemmer = lambda x: x
        else:
            raise (NotImplementedError)

        if type(sw_list) is str:
            self.sw_list = get_stopwords(sw_list)
        elif type(sw_list) is list:
            self.sw_list = sw_list
        elif sw_list is False:
            self.sw_list = []
        else:
            raise (NotImplementedError)

        self.id_mapping = None
        self.inverted_index = None
        self.doc_count = None
        self.doc_lens = None
        self.relative_doc_lens = None
        self.docs_path = None
        self.doc_index = None

    def index(
        self,
        collection: List[Dict],
        show_progress: bool = True,
    ):
        self.docs_path = str(self.index_path / "docs.jsonl")

        with open(self.docs_path, "wb") as f:
            for y in track(
                collection,
                disable=not show_progress,
                description="Saving collection",
            ):
                f.write(orjson.dumps(y) + "\n".encode())

        # write_jsonl(collection, self.docs_path)
        self.doc_index = Indxr(self.docs_path)

        self.id_mapping = {i: x["id"] for i, x in enumerate(collection)}
        self.doc_count = len(self.id_mapping)

        # Preprocessing --------------------------------------------------------
        collection = self.preprocessing(collection, show_progress)

        # Inverted index -------------------------------------------------------
        self.inverted_index = self.build_inverted_index(
            collection, show_progress
        )

        # Pre-compute some stuff -----------------------------------------------
        self.doc_lens = self.get_doc_lens(collection)
        self.relative_doc_lens = self.get_relative_doc_lens(self.doc_lens)

        self.save()

    def index_file(
        self,
        path: str,
        show_progress: bool = True,
        callback: Callable = None,
    ):
        kind = os.path.splitext(path)[1][1:]
        assert kind in {
            "jsonl",
            "csv",
            "tsv",
        }, "Only JSONl, CSV, and TSV are currently supported."

        if kind == "jsonl":
            collection = read_jsonl(path, callback=callback)
        elif kind == "csv":
            collection = read_csv(path, callback=callback)
        elif kind == "tsv":
            collection = read_csv(path, delimiter="\t", callback=callback)

        self.index(
            collection=collection,
            show_progress=show_progress,
        )

    def save(self):
        state = {
            "init_args": self.init_args,
            "docs_path": self.docs_path,
            "id_mapping": self.id_mapping,
            "doc_count": self.doc_count,
            "inverted_index": self.inverted_index,
            "doc_lens": self.doc_lens,
            "relative_doc_lens": self.relative_doc_lens,
        }

        np.save(self.index_path / "index.npy", state)

    @staticmethod
    def load(index_name="new-index"):
        index_path = home_path() / index_name / "index.npy"
        state = np.load(index_path, allow_pickle=True)[()]
        se = SearchEngine(**state["init_args"])
        se.docs_path = state["docs_path"]
        se.doc_index = Indxr(se.docs_path)
        se.id_mapping = state["id_mapping"]
        se.doc_count = state["doc_count"]
        se.inverted_index = state["inverted_index"]
        se.doc_lens = state["doc_lens"]
        se.relative_doc_lens = state["relative_doc_lens"]

        return se

    @staticmethod
    def delete(index_name="new-index"):
        try:
            shutil.rmtree(home_path() / index_name)
            print(f"{index_name} successfully removed.")
        except FileNotFoundError:
            print(f"{index_name} not found.")

    def preprocessing(self, collection, show_progress: bool = True):
        inputs = [
            (
                x["contents"],
                self.tokenizer,
                self.sw_list,
                self.stemmer,
            )
            for x in collection
        ]

        with Pool() as p:
            preprocessed = list(
                track(
                    p.imap(
                        prepro_caller,
                        inputs,
                        chunksize=1_000,
                    ),
                    total=self.doc_count,
                    disable=not show_progress,
                    description="Processing texts",
                )
            )

        for x, y in zip(collection, preprocessed):
            x["contents"] = y

        return collection

    def get_doc_lens(self, collection):
        return [len(x["contents"]) for x in collection]

    def get_relative_doc_lens(self, doc_lens):
        return (doc_lens / np.mean(doc_lens)).astype(np.float32)

    def build_inverted_index(self, collection, show_progress: bool = True):
        inverted_index = defaultdict(lambda: defaultdict(list))

        for i, doc in enumerate(
            track(
                collection,
                disable=not show_progress,
                description="Building inverted index",
            )
        ):
            unique, counts = np.unique(doc["contents"], return_counts=True)

            for term, occurrences in zip(unique, counts):
                inverted_index[term]["doc_ids"].append(i)
                inverted_index[term]["tfs"].append(occurrences)

        for term in track(
            inverted_index.keys(),
            total=len(inverted_index.keys()),
            disable=not show_progress,
            description="Optimizing inverted index",
        ):
            if len(inverted_index[term]["doc_ids"]) < self.min_term_freq:
                del inverted_index[term]

            inverted_index[term]["doc_ids"] = np.array(
                inverted_index[term]["doc_ids"], dtype=np.int32
            )
            inverted_index[term]["tfs"] = np.array(
                inverted_index[term]["tfs"], dtype=np.int16
            )
            inverted_index[term] = dict(inverted_index[term])

        return dict(inverted_index)

    def get_term_doc_freqs(self, query_terms: List[str]) -> nb.types.List:
        return TypedList(
            [
                self.inverted_index[t]["tfs"]
                for t in query_terms
                if t in self.inverted_index
            ]
        )

    def get_doc_ids(self, query_terms: List[str]) -> nb.types.List:
        return TypedList(
            [
                self.inverted_index[t]["doc_ids"]
                for t in query_terms
                if t in self.inverted_index
            ]
        )

    def search(
        self,
        query: str,
        return_docs: bool = True,
        b: float = 0.75,
        k1: float = 1.2,
        n_res: int = 100,
    ):
        query_terms = prepro(query, self.tokenizer, self.sw_list, self.stemmer)
        term_doc_freqs = self.get_term_doc_freqs(query_terms)
        doc_ids = self.get_doc_ids(query_terms)

        unique_doc_ids, scores = bm25(
            b=b,
            k1=k1,
            term_doc_freqs=term_doc_freqs,
            doc_ids=doc_ids,
            relative_doc_lens=self.relative_doc_lens,
            n_res=n_res,
        )

        unique_doc_ids = [self.id_mapping[doc_id] for doc_id in unique_doc_ids]

        if return_docs:
            docs = self.doc_index.mget(unique_doc_ids)
            results = []
            for doc, score in zip(docs, scores):
                doc["score"] = score
                results.append(doc)
            return results
        else:
            return dict(zip(unique_doc_ids, scores))

    def msearch(
        self,
        queries: Dict[str, str],
        b: float = 0.75,
        k1: float = 1.2,
        n_res: int = 100,
    ):

        term_doc_freqs = TypedList()
        doc_ids = TypedList()
        q_ids = []
        no_results_q_ids = []

        for q_id, query in queries.items():
            query_terms = prepro(
                query, self.tokenizer, self.sw_list, self.stemmer
            )

            if all(t not in self.inverted_index for t in query_terms):
                no_results_q_ids.append(q_id)
                continue

            q_ids.append(q_id)
            term_doc_freqs.append(self.get_term_doc_freqs(query_terms))
            doc_ids.append(self.get_doc_ids(query_terms))

        unique_doc_ids, scores = bm25_multi(
            b=b,
            k1=k1,
            term_doc_freqs=term_doc_freqs,
            doc_ids=doc_ids,
            relative_doc_lens=self.relative_doc_lens,
            n_res=n_res,
        )

        unique_doc_ids = [
            [self.id_mapping[doc_id] for doc_id in _unique_doc_ids]
            for _unique_doc_ids in unique_doc_ids
        ]

        results = {
            q: dict(zip(unique_doc_ids[i], scores[i]))
            for i, q in enumerate(q_ids)
        }

        for q_id in no_results_q_ids:
            results[q_id] = {}

        # Order as queries
        return {q_id: results[q_id] for q_id in queries}

    def autotune(
        self,
        queries: Dict[str, str],
        qrels: Dict[str, Dict[str, float]],
        metric: str = "ndcg@100",
        n_trials: int = 100,
        n_res: int = 100,
    ):
        return tune_bm25(
            queries=queries,
            qrels=qrels,
            se=self,
            metric=metric,
            n_trials=n_trials,
            n_res=n_res,
        )



from typing import Tuple

import numba as nb
import numpy as np
from numba import njit, prange
from numba.typed import List as TypedList


@njit(cache=True)
def concat1d(X):
    out = np.empty(sum([len(x) for x in X]), dtype=X[0].dtype)

    i = 0
    for x in X:
        for j in range(len(x)):
            out[i] = x[j]
            i = i + 1

    return out


@njit(cache=True)
def bm25(
    b: float,
    k1: float,
    term_doc_freqs: nb.typed.List[np.ndarray],
    doc_ids: nb.typed.List[np.ndarray],
    relative_doc_lens: nb.typed.List[np.ndarray],
    n_res: int,
) -> Tuple[np.ndarray]:
    unique_doc_ids = np.unique(concat1d(doc_ids))

    doc_count = len(relative_doc_lens)
    scores = np.empty(doc_count, dtype=np.float32)
    scores[unique_doc_ids] = 0.0  # Initialize scores

    for i in range(len(term_doc_freqs)):
        indices = doc_ids[i]
        freqs = term_doc_freqs[i]

        df = np.float32(len(indices))
        idf = np.float32(np.log(1.0 + (((doc_count - df) + 0.5) / (df + 0.5))))

        # BM25_score = IDF * (tf * (k + 1)) / (k * (1.0 - b + b * (|d|/avgDl)) + tf)
        scores[indices] += idf * (
            (freqs * (k1 + 1.0))
            / (freqs + k1 * (1.0 - b + (b * relative_doc_lens[indices])))
        )

    scores = scores[unique_doc_ids]
    indices = np.argsort(scores)[::-1][:n_res]

    return unique_doc_ids[indices], scores[indices]


@njit(parallel=True, cache=True)
def bm25_multi(
    b: float,
    k1: float,
    term_doc_freqs: nb.typed.List[nb.typed.List[np.ndarray]],
    doc_ids: nb.typed.List[nb.typed.List[np.ndarray]],
    relative_doc_lens: nb.typed.List[nb.typed.List[np.ndarray]],
    n_res: int,
) -> Tuple[nb.typed.List[np.ndarray]]:
    unique_doc_ids = TypedList([np.empty(1, dtype=np.int32) for _ in doc_ids])
    scores = TypedList([np.empty(1, dtype=np.float32) for _ in doc_ids])

    for i in prange(len(term_doc_freqs)):
        unique_doc_ids[i], scores[i] = bm25(
            b=b,
            k1=k1,
            term_doc_freqs=term_doc_freqs[i],
            doc_ids=doc_ids[i],
            relative_doc_lens=relative_doc_lens,
            n_res=n_res,
        )

    return unique_doc_ids, scores



from typing import Callable

from nltk.tokenize import sent_tokenize, word_tokenize, wordpunct_tokenize

tokenizers_dict = {
    "whitespace": str.split,
    "word": word_tokenize,
    "wordpunct": wordpunct_tokenize,
    "sent": sent_tokenize,
}


def get_tokenizer(tokenizer: str) -> Callable:
    assert (
        tokenizer.lower() in tokenizers_dict
    ), f"Tokenizer {tokenizer} not supported."
    return tokenizers_dict[tokenizer.lower()]



import warnings

import optuna
from optuna.exceptions import ExperimentalWarning
from ranx import Qrels, Run, evaluate

warnings.filterwarnings("ignore", category=ExperimentalWarning)


def bm25_objective(trial, queries, qrels, se, metric, n_res):
    b = trial.suggest_float("b", 0.0, 1.0, step=0.01)
    k1 = trial.suggest_float("k1", 0.0, 10.0, step=0.1)

    run = Run(se.msearch(queries=queries, b=b, k1=k1, n_res=n_res))

    return evaluate(qrels, run, metric)


def tune_bm25(queries, qrels, se, metric, n_trials, n_res):
    qrels = Qrels(qrels)

    optuna.logging.set_verbosity(optuna.logging.WARNING)

    sampler = optuna.samplers.TPESampler(seed=42)
    study = optuna.create_study(direction="maximize", sampler=sampler)
    study.optimize(
        lambda trial: bm25_objective(trial, queries, qrels, se, metric, n_res),
        n_trials=n_trials,
        show_progress_bar=True,
    )

    optuna.logging.set_verbosity(optuna.logging.INFO)

    return study.best_params



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="retriv",
    version="0.1.1",
    author="Elias Bassani",
    author_email="elias.bssn@gmail.com",
    description="retriv: A Blazing-Fast Python Search Engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AmenRa/retriv",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy",
        "nltk",
        "numba>=0.54.1",
        "rich",
        "optuna",
        "krovetzstemmer",
        "ranx",
        "indxr",
        "oneliner_utils",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: General",
    ],
    keywords=["information retrieval", "search engine", "bm25", "numba"],
    python_requires=">=3.7",
)



from typing import List

import nltk

nltk.download("stopwords", quiet=True)

supported_languages = {
    "arabic",
    "azerbaijani",
    "basque",
    "bengali",
    "catalan",
    "chinese",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "greek",
    "hebrew",
    "hinglish",
    "hungarian",
    "indonesian",
    "italian",
    "kazakh",
    "nepali",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "slovene",
    "spanish",
    "swedish",
    "tajik",
    "turkish",
}


def get_stopwords(lang: str) -> List:
    assert (
        lang.lower() in supported_languages
    ), f"Stop-words for {lang} are not available."
    return nltk.corpus.stopwords.words(lang)
